#ifndef GAMEOBJECT_INCLUDED
	#define GAMEOBJECT_INCLUDED
// gameobject.h
// abstract base class
#include "vector3d.h"
#include "map.h"

#include <iostream>

class Manager; // forward declaration to prevent recursion.
class HoverCraft;

class GameObject {
	protected:
		Vect3d location;
		Component radius;
		Manager *manager;
		bool visible;
		bool dead;
		bool collidable;
		bool solid;
	public:
		GameObject(Manager *m) : manager(m), visible(true), dead(false), collidable(true), solid(false) { }
		bool isVisible() { return visible; }
		bool isDead() { return dead; }
		void setDead(bool b) { dead = b; }
		bool isCollidable() { return collidable; }
		bool isSolid() { return solid; }
		virtual void setLocation(const Vect3d &where) { location = where; }
		Vect3d getLocation() { return location; }
		Component getRadius() { return radius; }
		virtual ~GameObject() { }
		virtual void renderTo(const View &v) = 0;
		virtual void tick(Component quantum) = 0;
		virtual char saveChar() { return 0; }
		// Save extra info if needed
		virtual void save(std::ostream &out) { }
		virtual void load(std::istream &in) { }
		virtual void handlePlayerCollision(HoverCraft *h) { }
};

// An item which sits there, doing nothing.
// When set to a particular location, attempts to plant itself on the ground.

class StaticObject : public GameObject {
	public:
		enum StaticObjectType { Flag = 0, SmallTree1, SmallTree2, Tree1, Tree2 };
	private:
		StaticObjectType type;
		BITMAP *spritebm;
		void setSprite();
		static int spriteRefs[];
	public:
		StaticObject(Manager *m) : GameObject(m) { solid = true; }
		StaticObject(Manager *m, StaticObjectType type);
		virtual ~StaticObject() { }
		virtual void renderTo(const View &v);
		virtual void tick(Component quantum) { }
		virtual void setLocation(const Vect3d &where);
		static const char SAVE_CHAR = 'S';
		virtual char saveChar() { return SAVE_CHAR; }
		virtual void save(std::ostream &out) { out << type; }
		virtual void load(std::istream &in) { int i; in >> i; type = (StaticObjectType) i; setSprite(); }
};

class EditorCursor : public GameObject {
	public:
		EditorCursor(Manager *m);
		virtual ~EditorCursor() { }
		virtual void renderTo(const View &v);
		virtual void tick(Component quantum);
		virtual void setLocation(const Vect3d &where);
	private:
		void plonkItem(GameObject *);
		void deleteItem();
		void snapItem();
		Component reloadTime;
		
};

// Checkpoint 0 is the start/finish line, others are around the course.

class CheckPoint : public GameObject {
	private:
		int number;
		void drawSpot(const View &v, Vect3d where);
		void setVectors();
		Component azimuth;
		Vect3d forward;
		Vect3d right;
	public:
		CheckPoint(Manager *m);
		virtual ~CheckPoint() { }
		virtual void renderTo(const View &v);
		virtual void tick(Component quantum) { }
		virtual void setLocation(const Vect3d &where);
		void setRadius(Component r) { radius = r; }
		void setAzimuth(Component a) { azimuth = a; setVectors(); }
		Component getAzimuth() { return azimuth; }
		Vect3d getForward() { return forward; }
		static const char SAVE_CHAR = 'C';
		virtual char saveChar() { return SAVE_CHAR; }
		virtual void save(std::ostream &out) { out << radius << " " << number << " " << azimuth; }
		virtual void load(std::istream &in);
		int getNumber() { return number; }
};

class Powerup: public GameObject {
	public:
		enum PowerupType { Freeze=1, OutOfControl=2, FreezeOpponents=3, Turbo=4 };
		Powerup(Manager *m, PowerupType t = Turbo);
		virtual ~Powerup() { }
		virtual void renderTo(const View &v);
		virtual void tick(Component quantum);
		static const char SAVE_CHAR = 'P';
		virtual char saveChar() { return SAVE_CHAR; }
		virtual void save(std::ostream &out) { out << (int) type; }
		virtual void load(std::istream &in) { int t; in >> t; type = (PowerupType) t; }
		void setType(PowerupType t) { type = t; }
		void handlePlayerCollision(HoverCraft *h);
	private:
		PowerupType type;
		Component animationTime;
		Component respawnTime;
	
};

#endif // GAMEOBJECT_INCLUDED
