#include "manager.h"
#include "constants.h"
#include "data.h"
#include "hovercraft.h"
#include "ai.h"
#include "language.h"

#include <stdexcept>
#include <iostream>
#include <algorithm>
#include <fstream>
#include <sstream>

static void showMessage(const char *msg)
{
	int green = makecol(0,120,0);
	int white = makecol(255,255,255);
	clear_to_color(screen, green);
	textout(screen, font, msg, 10,10, white);
}

static void deleteObjectList(Manager::GameObjectList &list) {
	for (Manager::GameObjectList::iterator i = list.begin(); i != list.end(); i++) {
		GameObject * obj = *i;
		delete obj;
	}
	list.clear();
}

Manager::Manager(bool edit)
{
	editMode = edit;
	backdrop = 0;
	buffer = 0;
	df = 0; // no datafile yet.
	trackObject = 0;
	playerObject = 0;
	backdrop = create_bitmap(VirtualWidth, VirtualHeight);
	if (backdrop == 0) {
		throw std::runtime_error("Cannot create virtual screen");
	}
	buffer = create_bitmap(ScreenWidth, ScreenHeight);
	if (buffer == 0) {
		destroy_bitmap(backdrop);
		throw std::runtime_error("Cannot create buffer screen");
	}
	timems = 0;
	timens = 0;
	messageTime = 0;
	musicOn = true;
}

Manager::~Manager()
{
	clearObjects();
	if (backdrop != 0)
		destroy_bitmap(backdrop);
	if (buffer != 0)
		destroy_bitmap(buffer);
	if (df != 0) 
		unload_datafile(df);
}

void Manager::loadData()
{
	//set_color_conversion(COLORCONV_24_TO_16 | COLORCONV_24_TO_32 | COLORCONV_KEEP_TRANS);
	set_color_conversion(COLORCONV_KEEP_TRANS);
	if (df == 0) {
		df = load_datafile("data.dat");
		if (df == 0) {
			throw std::runtime_error("Failed to load datafile");
		}
	}
}

void Manager::initLevel(int l)
{
	level = l;
	showMessage("Please wait... rendering map");
	loadData(); // load the datafile if not already there
	clearObjects(); // ditch any previous objects.
	// Create a view object
	view = View(Point(VirtualWidth / 2, VirtualHeight + 320), backdrop);
	// Load the map.
	map.load(l);
	// Render the map into the backdrop
	clear_to_color(backdrop,0);
	map.renderTo(view);
	offsetx = 0;
	offsety = 0;
	
	// load the objects 
	loadTrack(); // load the track object data

	if (editMode) {	
		// Enable edit mode...
		editMode = true;
		Component centre = (Map::BlocksWide * Map::BlockSize) / 2;
		EditorCursor *cursor = new EditorCursor(this);
		cursor->setLocation(Vect3d(centre, centre, 0));
		addObject(cursor);
		trackObject = cursor;
	} else {
		// Init vehicles...
		initVehicles();
	}
}

void Manager::initVehicles()
{
	// Run a short tick so that stuff can get itself settled...
	handleLogic(0.1);
	// Discover the first checkpoint..
	CheckPoint *firstCheckPoint = 0;
	for(GameObjectIterator i = gameObjects.begin(); i != gameObjects.end(); i++) {
		GameObject * obj = *i;
		CheckPoint * cp = dynamic_cast<CheckPoint *>(obj);
		if (cp) {
			if (cp->getNumber() == 0) {
				firstCheckPoint = cp;
				break;
			}
		}
	}
	if (firstCheckPoint == 0) {
		throw std::runtime_error("Cannot find the first checkpoint");
	}
	
	Vect3d gridPos = firstCheckPoint->getLocation();
	gridPos = gridPos - (firstCheckPoint->getForward() * 30);
	Vect3d gridOffset(firstCheckPoint->getForward().y,
		- firstCheckPoint->getForward().x, 0);
	gridOffset = gridOffset * 20;
	for (int i=0; i<4; i++) {
		HoverCraft *h = new HoverCraft(this,4 - i);
		h->setLocation(gridPos + gridOffset);
		h->setAzimuth(firstCheckPoint->getAzimuth());
		if (i == 3) {
			h->setController(new KeyboardController());
			playerObject = h;
			trackObject = h;
			h->setIsPlayer(true);
		} else {
			h->setController(new AiController(this, h));
		}
		addObject(h);
		gridOffset = - gridOffset;
		gridPos = gridPos - (firstCheckPoint->getForward() * 30);
	}
	numLaps = 3;
	sortPlaces();
}

void Manager::render()
{
	// Sort out the viewport
	if (trackObject) {
		Point p(view.translate(trackObject->getLocation()));
		p.x -= (ScreenWidth / 2);
		p.y -= (ScreenHeight / 2);
		offsetx = (int) p.x;
		offsety = (int) p.y;
		// TODO: clip
	}
	
	// Blit backdrop into the buffer
	blit(backdrop, buffer, offsetx,offsety,0,0, ScreenWidth, ScreenHeight);
	// Draw objects
	View bufferView = view;
	bufferView.bottomLeft.x -= offsetx;
	bufferView.bottomLeft.y -= offsety;
	bufferView.bitmap = buffer;
	// TODO:cull them (but not in that order)
	// Sort them by adding into a multimap with depth as a key.
	renderList.clear();
	for(GameObjectIterator i = gameObjects.begin(); i != gameObjects.end(); i++) {
		GameObject * obj = *i;
		Vect3d where = obj->getLocation();
		Point p(bufferView.translate(where));
		bool visible = obj->isVisible();
		// Perform culling.
		if (visible) {
			Component diameter = obj->getRadius() * 2;
			if ((p.x + diameter) < 0) visible = false;
			if ((p.x - diameter) > ScreenWidth) visible = false;
			if ((p.y + diameter) < 0) visible = false;
			// Take into account that the object could be fairly high
			if ((p.y - (diameter * 3)) > ScreenHeight) visible = false;
		}
		if (visible) {
			// Need to invert depth so we get furthest objects first.
			Component depth = - (where.x + where.y);
			renderList.insert(RenderList::value_type(depth, obj));
		}
	}
	int visObjects = 0;
	for(RenderList::iterator i = renderList.begin(); i != renderList.end(); i++) {
		RenderList::value_type pair = *i;
		GameObject * obj = pair.second;
		obj->renderTo(bufferView);
		visObjects ++;
	}
	if (playerObject) {
		playerObject->renderHud(buffer);
	}
	// Show the big message...
	if (messageTime > 0) {
		int n = (int) (messageTime * 255.0);
		if (n>255) n = 255;
		int messageColour = makecol(n,n,n);
		FONT *font = (FONT *) df[INTREPID_PCX].dat;
		text_mode(-1);
		textout(buffer, font, bigMessage.c_str(), 140, 200, messageColour);
	}
	
	// Blit buffer into screen.
	blit(buffer, screen, 0,0,0,0, ScreenWidth, ScreenHeight);
}

static bool isDead(GameObject *obj) {
	return obj->isDead();
}

void Manager::handleCollision(HoverCraft *veh, GameObject *obj)
{
	veh->handleCollision(obj);
}

void Manager::findCollisions()
{
	// For each vehicle
	for (VehicleList::iterator i = vehicles.begin(); i != vehicles.end(); i++) {
		// Test it with each other object which wants collisions
		HoverCraft *veh = *i;
		Vect3d vehiclePos(veh->getLocation());
		for(GameObjectIterator j = gameObjects.begin(); j != gameObjects.end(); j++) {
			GameObject * obj = *j;
			// don't collide with yourself or non-collidable things
			if ((veh != obj) && (obj->isCollidable())) { 
				Vect3d objPos(obj->getLocation());
				// And when it collides, call the appropriate method on the vehicle
				// To register a collision
				Component distSquared = (objPos - vehiclePos).magnitudeSquared();
				Component radiiSquared = (obj->getRadius() + veh->getRadius());
				radiiSquared = radiiSquared * radiiSquared ;
				if (distSquared < radiiSquared) {
					// But when two vehicles collide, handle that differently.
					// "Collide" doesn't necessarily mean a collision, it applies to checkpoints
					// and AI waypoints too
					handleCollision(veh, obj);
					obj->handlePlayerCollision(veh);
				}
			}
		}
	}
}

void Manager::handleLogic(Component quantum)
{
	// Dump the newObjects into the list...
	for(GameObjectIterator i = newObjects.begin(); i != newObjects.end(); i++) {
		GameObject * obj = *i;
		gameObjects.push_back(obj);
	}
	newObjects.clear();
	// Run logic
	for(GameObjectIterator i = gameObjects.begin(); i != gameObjects.end(); i++) {
		GameObject * obj = *i;
		obj->tick(quantum);
	}
	// Handle collisions
	findCollisions();
	// Free any old objects from last time..
	deleteObjectList(deadList);
	// Now trash any dead ones...
	for(GameObjectIterator i = gameObjects.begin(); i != gameObjects.end(); i++) {
		GameObject * obj = *i;
		if (obj->isDead()) deadList.push_back(obj);
	}
	// Wipe them out.
	GameObjectIterator new_end =
		std::remove_if(gameObjects.begin(), gameObjects.end(),isDead);
	gameObjects.erase(
		new_end,
		gameObjects.end());
	// In case we delete our own self...
	if (trackObject) {
		if (trackObject->isDead()) {
			trackObject = 0;
		}
	}
	// Add some time...
	timens += (long) (quantum * 1E9);
	timems += (timens / 1000000);
	timens = timens % 1000000;
	messageTime -= quantum;
}

void Manager::loadTrack()
{
	using namespace std;
	ostringstream oss;
        oss << "tracks/track" << level << ".txt";
	ifstream infile(oss.str().c_str());
	bool finished = false;
	numCheckpoints = 0;
	do {
		char c;
		infile >> c;
		GameObject *obj = 0;
		switch (c) {
			case 'E': // End marker
				finished = true;
				break;
			case StaticObject::SAVE_CHAR: 
				obj = new StaticObject(this);
				break;
			case CheckPoint::SAVE_CHAR:
				obj = new CheckPoint(this);
				break;
			case AiWaypoint::SAVE_CHAR:
				{
					AiWaypoint *wp = new AiWaypoint(this);
					waypoints.push_back(wp);
					obj = wp;
					break;
				}
			case Powerup::SAVE_CHAR:
				obj = new Powerup(this);
				break;
		}
		if (obj) {
			Vect3d location;
			infile >> location.x;
			infile >> location.y;
			infile >> location.z;
			obj->setLocation(location);
			obj->load(infile);
			addObject(obj);
		}
	} while ((! finished) && (infile.good()));
	std::cout << "NUmber of checkpoints:" << numCheckpoints << std::endl;
}

void Manager::saveTrack()
{
	using namespace std;
	ofstream out("newtrack.txt");
	for(GameObjectIterator i = gameObjects.begin(); i != gameObjects.end(); i++) {
		GameObject * obj = *i;
		char saveChar = obj->saveChar();
		Vect3d where = obj->getLocation();
		if (saveChar) {
			out << saveChar << " " << 
				where.x << " " << where.y << " " << where.z;
			out << " ";
			obj->save(out);
			out << endl;
		}
	}
	out << "E" << endl;
}

static volatile bool timerready;
static void timerfunc()
{
        timerready = true;
}

void Manager::mainLoop()
{
	Component quantum = 0;
	showBigMessage(language::current->Go);
	if (musicOn)
		play_midi((MIDI *) df[PIZZICATO_MID].dat, false);
	install_int(timerfunc,1000);
	int frames = 0;
	bool paused = false;
	do {
		if (key[KEY_PAUSE]) {
			// pause...
			paused = true;
		}
		if ((quantum > 0) && (! paused)) {
			handleLogic(quantum);
		}
		if (key[KEY_SPACE] && paused) {
			// Recal timer.
			remove_int(timerfunc);
			frames = 0;
			install_int(timerfunc,1000);
		}
		if (timerready) {
			// calibrate quantum...
			quantum = (1.0 / (Component) frames);
			timerready = false;
			frames = 0;
		}
		render();
		frames ++;
	} while (! key[KEY_ESC]);
	remove_int(timerfunc);
	stop_midi();
		
	if (editMode) {
		// Save changes...
		saveTrack();
	}
}

void Manager::preview(int l)
{
	level = l;
	showMessage("Generating preview");	
	clear_to_color(backdrop,0);
	
	loadData(); // load the datafile if not already there
	clearObjects(); // ditch any previous objects.
	// Create a view object
	view = View(Point(VirtualWidth / 2, VirtualHeight + 320), backdrop);
	// Load the map.
	map.load(l);
	// Render the map into the backdrop
	clear_to_color(backdrop,0);
	
	// load the objects 
	loadTrack(); // load the track object data
	handleLogic(0.1); // make sure the objects are seen to really exist...
	
	map.renderTo(view);
	for(GameObjectIterator i = gameObjects.begin(); i != gameObjects.end(); i++) {
		GameObject * obj = *i;
		bool visible = obj->isVisible();
		if (visible) {
			obj->renderTo(view);
		}
	}
	clear_to_color(screen,0);
	stretch_blit(backdrop, screen, 0,0, VirtualWidth, VirtualHeight, 0,0, 640, 320);
	/*BITMAP *fancy = create_bitmap(800,400);
	stretch_blit(backdrop, fancy, 0,0, VirtualWidth, VirtualHeight, 0,0, 800, 400);
	 save_bitmap("shot.tga", fancy, 0);
	 destroy_bitmap(fancy);
	*/
	
	readkey();
}


BITMAP * Manager::getSprite(int n)
{
	loadData();
	return (BITMAP *) df[n].dat;
}

SAMPLE * Manager::getSample(int n)
{
	loadData();
	return (SAMPLE *) df[n].dat;
}

void Manager::addObject(HoverCraft *obj)
{
	newObjects.push_back(obj);
	vehicles.push_back(obj);
}

void Manager::addObject(GameObject *obj)
{
	newObjects.push_back(obj);
}

void Manager::clearObjects()
{
	deleteObjectList(gameObjects);
	deleteObjectList(newObjects);
	deleteObjectList(deadList);
	vehicles.clear();
	waypoints.clear();
	trackObject = 0;
}

GameObject *Manager::findNearest(GameObject *me)
{
	Component minDistanceSquared = 1000000.0 * 1000000.0;
	GameObject *target = 0;
	Vect3d where(me->getLocation());
	for(GameObjectIterator i = gameObjects.begin(); i != gameObjects.end(); i++) {
		GameObject * obj = *i;
		if (obj != me) {
			Vect3d relative = obj->getLocation() - where;
			if (minDistanceSquared > (relative.magnitudeSquared())) {
				target = obj;
				minDistanceSquared = relative.magnitudeSquared();
			}
		}
	}
	return target;
}

// Returns TRUE if h1 is beating h2...
static bool RankComparator(HoverCraft *h1, HoverCraft *h2)
{
	int l1 = h1->getCurrentLap();
	int l2 = h2->getCurrentLap();
	if (l1 == l2) {
		int cp1 = h1->getLastCheckpoint();
		int cp2 = h2->getLastCheckpoint();
		if (cp1 == cp2) {
			return (h1->getLastCheckpointTime()) < (h2->getLastCheckpointTime());
		} else {
			return cp1 > cp2;
		}
	} else {
		return (l1 > l2);
	}
}

// Called whenever someone passes a checkpoint...
void Manager::sortPlaces()
{
	sort(vehicles.begin(), vehicles.end(), RankComparator);
	for (unsigned int i=0; i<vehicles.size(); i++) {
		vehicles[i]->setRank(i+1);	
	}
}

void Manager::freezeOpponents(HoverCraft *h)
{
	for (VehicleList::iterator i = vehicles.begin(); i != vehicles.end(); i++) {
		HoverCraft *other = *i;
		if (other != h) {
			other->freeze();
		}
	}
	if (h == playerObject)
	{
		showBigMessage(language::current->OpponentsFrozen);
		play_sample(getSample(TING_WAV), 192, 128, 1000, 0);
	}
}

void Manager::showBigMessage(const std::string &msg)
{
	bigMessage = msg;
	messageTime = 5.0;
}


