// map.cpp

#include "map.h"
#include <stdexcept>
#include <iostream>
#include <sstream>
#include <math.h>

// View methods

Point View::translate(const Vect3d &v) const {
	Point target(bottomLeft);
	target.x += v.x; // Xcoordinate makes x+1, y-0.5
	target.x -= v.y; // Ycoordinate makes x-1, y-0.5
	target.y -= (v.x + v.y) / 2;
	target.y -= v.z; // Z just decreases y
	return target;
};

// Map methods
Component Map::getHeight(const Point & p)
{
	int i = (int) (p.x / (Component) BlockSize);
	int j = (int) (p.y / (Component) BlockSize);
	if ((i<(BlocksWide-1)) && (j<(BlocksWide-1)) && (i>=0) && (j>=0)) {
		Component h1 = heightarray[i][j];
		Component h2 = heightarray[i+1][j];
		Component h3 = heightarray[i][j+1];
		Component h4 = heightarray[i+1][j+1];
		//interpolate.
		// These will be the weights of the x and y components respectively,
		// 0 means it's bang on [i][j] and 1 means it's [i+1][j+1]
		// Somewhere between 0 and 1
		Component xweight = fmod(p.x, ((Component) BlockSize)) / ((Component) BlockSize); 
		Component yweight = fmod(p.y, ((Component) BlockSize)) / ((Component) BlockSize); 
		// Find the weighted mean on the left and right of the cell
		Component left = (h1 * (1-yweight)) + (h3 * yweight);
		Component right = (h2 * (1-yweight)) + (h4 * yweight);
		// Return the weighted mean of those two values
		return (left * (1.0 - xweight)) + (right * xweight);
	} else {
		return 0;
	}
}

Vect3d Map::getNormal(const Point & p)
{
	// interpolate.
	int i = (int) (p.x / (Component) BlockSize);
	int j = (int) (p.y / (Component) BlockSize);
	if ((i<(BlocksWide-1)) && (j<(BlocksWide-1)) && (i>=0) && (j>=0)) {
		Vect3d n1 = normalarray[i][j];
		Vect3d n2 = normalarray[i+1][j];
		Vect3d n3 = normalarray[i][j+1];
		Vect3d n4 = normalarray[i+1][j+1];
		//interpolate.
		// These will be the weights of the x and y components respectively,
		// 0 means it's bang on [i][j] and 1 means it's [i+1][j+1]
		// Somewhere between 0 and 1
		Component xweight = fmod(p.x, ((Component) BlockSize)) / ((Component) BlockSize); 
		Component yweight = fmod(p.y, ((Component) BlockSize)) / ((Component) BlockSize); 
		// Find the weighted mean on the left and right of the cell
		Vect3d left = (n1 * (1-yweight)) + (n3 * yweight);
		Vect3d right = (n2 * (1-yweight)) + (n4 * yweight);
		// Return the weighted mean of those two values
		return (left * (1.0 - xweight)) + (right * xweight);
	} else {
		return Vect3d(0,0,1);
	}
}

void Map::load(int level) // TODO: specify which level to load
{
	using namespace std;
	// clear the old bitmaps out from a previous level
	destroyBitmaps();
	set_color_conversion(COLORCONV_TOTAL);
	// Load it once, set the palette, THEN load it again...
	df = load_datafile("tracks.dat");
	RGB *pal = (RGB *) df[1].dat;
	set_palette(pal);
	// This is a hack...
	if (df != 0) {
		unload_datafile(df);
	}
	df = load_datafile("tracks.dat");
	if (df == 0) {
		throw runtime_error("Unable to load tracks datafile");
	}
	
	unsigned char* heightDataPointer = (unsigned char*) df[0].dat;
	const int MapBytes = (BlocksWide * BlocksWide);
	colourmap = (BITMAP *) df[(level -1) + 2].dat;
	for (int i=0; i<BlocksWide; i++) {
		for (int j=0; j<BlocksWide; j++) {
			heightarray[i][j] = (Component) heightDataPointer[(MapBytes * (level-1)) + (j*BlocksWide)+ i];
		}
	}
	calculateNormals();
}

static Vect3d calculateNormal(const Vect3d &a, const Vect3d &b, const Vect3d &c)
{
	// Work out normal...
	Vect3d normal = (b-a) * (c-b);
	normal.normalise();
	return normal;
}

Vect3d Map::getPoint(int i, int j)
{
	Component height = heightarray[i][j];
	return Vect3d(i*BlockSize, j*BlockSize,  height);
}


void Map::calculateNormals()
{
	for (int i=1; i<BlocksWide-1; i++) {
		for (int j=1; j<BlocksWide-1; j++) {
			Vect3d centre = getPoint(i,j);
			Vect3d a = getPoint(i-1,j);
			Vect3d b = getPoint(i,j-1);
			Vect3d c = getPoint(i+1,j);
			Vect3d d = getPoint(i,j+1);
			// Find the normal of these four triangles, 
			// Then average them out, and normalise the result.
			Vect3d n1 = calculateNormal(centre,a,b);
			Vect3d n2 = calculateNormal(centre,b,c);
			Vect3d n3 = calculateNormal(centre,c,d);
			Vect3d n4 = calculateNormal(centre,d,a);
			Vect3d normal = (n1 + n2 + n3 + n4) * (1.0 / 4.0);
			normal.normalise();
			normalarray[i][j] = normal;
		}
	}
}

static Vect3d lightVector(-1,0.5,-1);

int Map::getColour(int i, int j)
{
	Vect3d normal = normalarray[i][j];
	Component highlight = normal.dotProduct(- lightVector);
	if (highlight < 0) highlight = 0;
	
	if (highlight > 1) highlight = 1;
	Component ambient = 0.4;
	Component diffuse = 1 - ambient;
	Component brightness = ambient + diffuse * highlight;
	int mapcolour = getpixel(colourmap,i,j);
	Component r = (Component) getr(mapcolour) * brightness;
	Component g = (Component) getg(mapcolour) * brightness;
	Component b = (Component) getb(mapcolour) * brightness;
	int colour = makecol((int) r, (int) g, (int) b);
	return colour;
}

bool Map::isWater(const Point &p)
{
	int i = (int) round(p.x / (Component) BlockSize);
	int j = (int) round(p.y / (Component) BlockSize);
	if (i<0) i=0;
	if (j<0) j=0;
	if (i>(BlocksWide-1)) return false;	
	if (j>(BlocksWide-1)) return false;
	int mapcolour = getpixel(colourmap,i,j);
	int r = getr(mapcolour);
	int g = getg(mapcolour);
	int b = getb(mapcolour);
	return (((b - r) > 0) && ((b-g) > 0));
}


static void triangle3d(BITMAP *bm, const View &v, 
	const Vect3d &a, const Vect3d &b, const Vect3d &c,
	int coloura, int colourb, int colourc)
{
	Point ascreen(v.translate(a));
	Point bscreen(v.translate(b));
	Point cscreen(v.translate(c));
	struct V3D_f points[3];
	points[0].x = ascreen.x;
	points[0].y = ascreen.y;
	points[0].c = coloura;
	points[1].x = bscreen.x;
	points[1].y = bscreen.y;
	points[1].c = colourb;
	points[2].x = cscreen.x;
	points[2].y = cscreen.y;
	points[2].c = colourc;
	// Not sure why i have to do this:
	struct V3D_f *ppoints[3];
	ppoints[0] = &points[0];
	ppoints[1] = &points[1];
	ppoints[2] = &points[2];
	
	polygon3d_f(bm, POLYTYPE_GCOL, 0 /* texure */, 3, ppoints);
} 

void Map::renderTo(const View &v)
{
	lightVector.normalise();
	for (int i=BlocksWide-2; i>0; i--) {
		for (int j=1; j<BlocksWide-1; j++) {
			Vect3d a = getPoint(i,j);
			int coloura = getColour(i,j);
			Vect3d b = getPoint(i,j-1);
			int colourb = getColour(i,j-1);
			Vect3d c = getPoint(i-1,j);
			int colourc = getColour(i-1,j);
			Vect3d d = getPoint(i-1,j-1);
			int colourd = getColour(i-1,j-1);
			// draw triangle abc
			triangle3d(v.bitmap, v, a, b, c, coloura, colourb, colourc);
			// draw triangle dcb
			triangle3d(v.bitmap, v, d, c, b, colourd, colourc, colourb);
		}
	}
}

Map::Map() {
	heightmap = 0;
	colourmap = 0;
	df = 0;
	// Init the height array
	heightarray = new Component *[BlocksWide];
	for (int i=0; i<BlocksWide; i++) {
		heightarray[i] = new Component[BlocksWide];
	}
	// Normals...
	normalarray = new Vect3d *[BlocksWide];
	for (int i=0; i<BlocksWide; i++) {
		normalarray[i] = new Vect3d[BlocksWide];
	}
}

void Map::destroyBitmaps()
{
	heightmap = 0;
	colourmap = 0;
	if (df != 0) {
		unload_datafile(df);
		df = 0;
	}
}

Map::~Map() {
	destroyBitmaps();
	for (int i=0; i<BlocksWide; i++) {
		delete [] heightarray[i];
		delete [] normalarray[i];
	}
	delete [] heightarray;
	delete [] normalarray;
	heightarray = 0;
}

